import unittest
from unittest.mock import patch

from mmisdk.custodian_factory import CustodianFactory
from mmisdk.qredo.qredo import Qredo
from mmisdk.utils.mock_response import MockResponse

MOCK_CUSTODIANS_CONFIG = {
    "portfolio": {
        "enabled": True,
        "url": "https://metamask-institutional.io/",
        "cookieSetUrls": [
            "https://mmi-portfolio-poc.codefi.network/api/update-dashboard-cookie",
            "https://metamask-institutional.io/api/update-dashboard-cookie"
        ]
    },
    "custodians": [
        {
            "refreshTokenUrl": None,
            "name": "cactus",
            "displayName": None,
            "enabled": None,
            "mmiApiUrl": "https://mmi.codefi.network/v1",
            "apiBaseUrl": None,
            "iconUrl": None,
            "isNoteToTraderSupported": False
        },
        {
            "refreshTokenUrl": None,
            "name": "qredo",
            "displayName": None,
            "enabled": None,
            "mmiApiUrl": "https://mmi.codefi.network/v1",
            "apiBaseUrl": None,
            "iconUrl": None,
            "isNoteToTraderSupported": True
        },
    ]
}


def mocked_requests_get(*args, **kwargs):
    """ This method will be used by the mock to replace requests.get"""

    # Mock the request that gets the custodian configs
    if args[0].__contains__('/configuration/default'):
        return MockResponse(MOCK_CUSTODIANS_CONFIG, 200)

    return MockResponse(None, 404)


class CustodianFactoryTest(unittest.TestCase):

    @patch('requests.get', side_effect=mocked_requests_get)
    def test_should_be_defined(self, mock_get):
        factory = CustodianFactory()
        self.assertIsNotNone(factory)

    @patch('requests.get', side_effect=mocked_requests_get)
    def test_should_fetch_configs(self, mock_requests):
        factory = CustodianFactory()
        self.assertIsNotNone(factory.custodians_config)
        self.assertGreater(len(factory.custodians_config), 0)

    @patch('requests.get', side_effect=mocked_requests_get)
    def test_should_create_custodian(self, mock_requests):
        factory = CustodianFactory()
        custodian = factory.create_for("qredo", "refresh_token")
        self.assertIsNotNone(custodian)
        self.assertIsInstance(custodian, Qredo)
        self.assertEqual(custodian.name, "qredo")
        self.assertEqual(custodian.api_url, "https://7ba211-api.qredo.net")
        self.assertEqual(custodian.refresh_token, "refresh_token")

    @patch('requests.get', side_effect=mocked_requests_get)
    def test_should_fail_unsupported_custodian(self, mock_requests):
        factory = CustodianFactory()
        with self.assertRaises(AssertionError):
            factory.create_for("whatever", "refresh_token")


if __name__ == "__main__":
    unittest.main()



from typing import Optional

from pydantic import BaseModel


class BaseTxParams(BaseModel):
    from_: str   # In Python, "from" is a reserved keyword
    to: str
    value: Optional[str]
    data: Optional[str]
    gasLimit: str
    gas: Optional[str]
    nonce: Optional[str]


class LegacyTxParams(BaseTxParams):
    type: Optional[str]
    gasPrice: str


class EIP11559TxParams(BaseTxParams):
    type: str
    maxPriorityFeePerGas: str
    maxFeePerGas: str



import requests

from mmisdk.common.custodian import Custodian
from mmisdk.common.transaction import Transaction
from mmisdk.common.transaction_params import EIP11559TxParams
from mmisdk.common.transaction_params import LegacyTxParams
from mmisdk.qredo.map_transaction_status import map_transaction_status
from mmisdk.qredo.qredo_tansaction import QredoTransaction
from mmisdk.qredo.qredo_tx_details import QredoTxDetails


class Qredo(Custodian):

    def __create_access_token(self):
        url = self.api_url+'/connect/token'
        querystring = {"grant_type": "refresh_token",
                       "refresh_token": self.refresh_token}
        payload = ""
        response = requests.post(url, data=payload, params=querystring)

        access_token = response.json()["access_token"]
        return access_token

    def __get_headers(self):
        """Returns the HTTP header to use in requests to the custodian"""
        access_token = self.__create_access_token()
        return {
            "Content-Type": "application/json",
            "Authorization": 'Bearer ' + access_token
        }

    def get_transaction(self, transaction_id):
        url = self.api_url + '/connect/transaction/' + transaction_id
        payload = ""
        headers = self.__get_headers()

        response = requests.get(url, data=payload, headers=headers)
        if (response.status_code != 200):
            raise requests.HTTPError(
                f"Couldn't get the transaction, {response.text}")

        # Parse response
        response_json = response.json()
        response_json["from_"] = response.json()["from"]
        qredo_transaction = QredoTransaction(**response_json)

        return self.__map_qredo_transaction_to_transaction(qredo_transaction)

    def create_transaction(self, qredo_tx_details, tx_params):
        url = self.api_url + "/connect/transaction"
        headers = self.__get_headers()

        # Parse and type check params
        qredo_tx_details_parsed = QredoTxDetails(**qredo_tx_details)
        tx_params_parsed = EIP11559TxParams(
            **tx_params) if 'type' in tx_params and tx_params['type'] == "2" else LegacyTxParams(**tx_params)

        # Build the payload
        type_ = tx_params_parsed.type  # "type" would shadow a builtin
        payload = {
            'to': tx_params_parsed.to,
            'from': tx_params_parsed.from_,
            'value': tx_params_parsed.value,
            'gasLimit': tx_params_parsed.gasLimit,
            'gasPrice': tx_params_parsed.gasPrice if type_ != "2" else None,
            'maxPriorityFeePerGas': tx_params_parsed.maxPriorityFeePerGas if type_ == "2" else None,
            'maxFeePerGas': tx_params_parsed.maxFeePerGas if type_ == "2" else None,
            'data': tx_params_parsed.data,
            'nonce': tx_params_parsed.nonce,
            'chainId': qredo_tx_details_parsed.chainId,
            'note': qredo_tx_details_parsed.note,
        }

        response = requests.post(url, json=payload, headers=headers)

        if (response.status_code != 200):
            raise requests.HTTPError(
                f"Couldn't create the transaction, {response.text}")

        # Parse response
        response_json = response.json()
        response_json["from_"] = response.json()["from"]
        qredo_transaction = QredoTransaction(**response_json)

        return self.__map_qredo_transaction_to_transaction(qredo_transaction)

    def __map_qredo_transaction_to_transaction(self, qredo_transaction: QredoTransaction) -> Transaction:
        return Transaction(
            id=qredo_transaction.txID,
            type="1" if qredo_transaction.gasPrice is not None else "2",
            from_=qredo_transaction.from_,
            to=qredo_transaction.to,
            value=qredo_transaction.value,
            gas=qredo_transaction.gasLimit,
            gasPrice=qredo_transaction.gasPrice,
            maxPriorityFeePerGas=None,
            maxFeePerGas=None,
            nonce=qredo_transaction.nonce,
            data=qredo_transaction.data,
            hash=qredo_transaction.txHash,
            status=map_transaction_status(qredo_transaction.status, "Unknown"),
        )



from mmisdk import CustodianFactory
from mmisdk import Qredo

# You can instanciate a custodian directly like so:
custodian0 = Qredo("qredo", "https://7ba211-api.qredo.net", "YOUR-REFRESH-TOKEN")
response0 = custodian0.get_transaction("1yiq2w7NSFei1uyQOremLiwbP2I")
print(response0.json())

# However it's simpler to rely on the factory
# It figures out itself which api URL to use
factory = CustodianFactory()
custodian1 = factory.create_for("qredo", "2ECrQyuycg6l4kjRGcnXvHzr87y-20984763496da41fc582ad946742f019")
response1 = custodian1.get_transaction("1yiq2w7NSFei1uyQOremLiwbP2I")
print(response1.json())

# {
#   "id": "1yiq2w7NSFei1uyQOremLiwbP2I",
#   "type": "1",
#   "from_": "0x6cF8426Cf36B1f43ffF6C376d290114297441bc2",
#   "to": "0x722dd3F80BAC40c951b51BdD28Dd19d435762180",
#   "value": "0",
#   "gas": "34062",
#   "gasPrice": "1000000010",
#   "maxPriorityFeePerGas": null,
#   "maxFeePerGas": null,
#   "nonce": "3",
#   "data": "0x97c5ed1e....",
#   "hash": "0xf31f4aa0d29e9d52a0f6de6dbb191f6060fd8bc994cea5c8785b2a94301a140c",
#   "status": {
#     "finished": false,
#     "submitted": true,
#     "signed": true,
#     "success": false,
#     "displayText": "Confirmed",
#     "reason": "Unknown"
#   }
# }



from typing import Optional

from pydantic import BaseModel


class TransactionStatus(BaseModel):
    finished: bool
    submitted: bool
    signed: bool
    success: bool
    displayText: str
    reason: str


class Transaction(BaseModel):
    """Common data format returned by all custodian clients."""
    id: str
    type: str
    from_: str
    to: str
    value: str
    gas: str
    gasPrice: Optional[str]
    maxPriorityFeePerGas: Optional[str]
    maxFeePerGas: Optional[str]
    nonce: str
    data: str
    hash: str
    status: TransactionStatus



import requests

from mmisdk.common.custodian import Custodian
from mmisdk.qredo.qredo import Qredo

MMI_CONFIGURATION_API = "https://mmi-configuration-api.codefi.network/v1/configuration/default"


class CustodianFactory:
    """Use this utility class to instanciate custodians."""

    def __init__(self) -> None:
        try:
            response = requests.get(MMI_CONFIGURATION_API, timeout=1)
            custodians_config = response.json()["custodians"]
        except TimeoutError:
            print('The request timed out')
        else:
            self.custodians_config = custodians_config

    def create_for(self, custodian_name, refresh_token) -> Custodian:
        """"""
        configs_with_name = filter(
            lambda config: config["name"] == custodian_name, self.custodians_config)

        as_list = list(configs_with_name)

        assert len(
            as_list) > 0, f"Could not find a custodian with name {custodian_name}"
        assert len(
            as_list) < 2, f"Found multiple custodians with name {custodian_name}"

        result = as_list[0]

        # TODO Add more cases
        if custodian_name == 'qredo':
            # TODO Deal with this
            result["apiBaseUrl"] = "https://7ba211-api.qredo.net"
            return Qredo(result["name"], result["apiBaseUrl"], refresh_token)
        else:
            assert False, "Passed custodian client is not implemented"



import unittest
from unittest import mock

from mmisdk.qredo.qredo import Qredo
from mmisdk.utils.mock_response import MockResponse

MOCK_TRANSACTION = {
    "txID": "2ELuLA4HWIzPBQB1MSLmbOxtoB1",
    "txHash": "0x6b5dc3519b325be4e6f9512f8d3b1fdb7af22a996e3f6bb1a898a28ef05c1963",
    "status": "authorized",
    "timestamps": {
        "authorized": 1662387127,
        "created": 1662387126
    },
    "events": [{
        "id": "2ELuLAcfbNcFqCmYPHlaGlkIdQf",
        "timestamp": 1662387126,
        "status": "created",
        "message": ""
    },
        {
        "id": "2ELuLBuOP6tc1Pp09FSHKvfvOoZ",
        "timestamp": 1662387127,
        "status": "authorized",
        "message": ""
    }],
    "nonce": 0,
    "gasPrice": "1000",
    "gasLimit": "21000",
    "from": "0x62468FD916bF27A3b76d3de2e5280e53078e13E1",
    "to":   "0x9999999999999999999999999999999999999999",
    "value": "1",
    "data": "",
    "rawTX": "4YCCA-iCUgiUYkaP2Ra_J6O3bT3i5SgOUweOE-EBgICAgA",
    "createdBy": "EFvSMt9uGTEsDCx22sYB8BPz1bCyxAZNPKDWpJVsLKM9",
    "accountID": "2ECrNK9dUlYIpdI4xVuRQ4Diuwq",
    "network": "",
    "chainID": "1"
}


def mocked_requests_get(*args, **kwargs):
    """ This method will be used by the mock to replace requests.get"""

    # Mock the request that gets the transaction
    if args[0].__contains__('/connect/transaction'):
        return MockResponse(MOCK_TRANSACTION, 200)

    return MockResponse(None, 404)


def mocked_requests_post(*args, **kwargs):
    """ This method will be used by the mock to replace requests.post"""

    # Mock the request that creates the access token
    if args[0].__contains__('/connect/token'):
        return MockResponse({"access_token": "some-access-token"}, 200)

    # Mock the request that creates a transaction
    if args[0].__contains__('/connect/transaction'):
        return MockResponse(MOCK_TRANSACTION, 200)

    return MockResponse(None, 404)


class QredoTest(unittest.TestCase):
    """Test Class"""
    @classmethod
    def setUpClass(cls):
        cls.custodian = Qredo("qredo", "http://some-url", "some-refresh-token")

    def test_should_be_defined(self):
        self.assertIsNotNone(self.custodian)

    @mock.patch('requests.get', side_effect=mocked_requests_get)
    @mock.patch('requests.post', side_effect=mocked_requests_post)
    def test_should_get_transaction(self, mock_get, mock_post):
        """We mock the methods requests.get and requests.post via the above decorators"""

        transaction = self.custodian.get_transaction(
            "2ELuLA4HWIzPBQB1MSLmbOxtoB1")
        self.assertEqual(transaction.id, '2ELuLA4HWIzPBQB1MSLmbOxtoB1')
        self.assertEqual(transaction.type, '1')
        self.assertEqual(transaction.from_,
                         '0x62468FD916bF27A3b76d3de2e5280e53078e13E1')
        self.assertEqual(
            transaction.to, '0x9999999999999999999999999999999999999999')
        self.assertEqual(transaction.value, '1')
        self.assertEqual(transaction.gas, '21000')
        self.assertEqual(transaction.gasPrice, '1000')
        self.assertEqual(transaction.maxPriorityFeePerGas, None)
        self.assertEqual(transaction.maxFeePerGas, None)
        self.assertEqual(transaction.nonce, '0')
        self.assertEqual(
            transaction.hash, '0x6b5dc3519b325be4e6f9512f8d3b1fdb7af22a996e3f6bb1a898a28ef05c1963')
        self.assertEqual(transaction.status, {
            "finished": False,
            "submitted": False,
            "signed": False,
            "success": False,
            "displayText": 'Authorized',
            "reason": "Unknown",
        })

    @mock.patch('requests.post', side_effect=mocked_requests_post)
    def test_should_create_transaction(self, mock_post):
        """We mock the methods requests.get and requests.post via the above decorators"""
        qredo_tx_details = {
            "chainId": "1",
            "note": "Some note"
        }
        tx_params = {
            "from_": "0x62468FD916bF27A3b76d3de2e5280e53078e13E1",
            "to": "0x9999999999999999999999999999999999999999",
            "value": "1",
            "data": "",
            "gasLimit": "21000",
            "gas": "1000",
            "nonce": 0,
            "type": "1",
            "gasPrice": "1000"
        }
        transaction = self.custodian.create_transaction(
            qredo_tx_details, tx_params)
        self.assertEqual(transaction.id, '2ELuLA4HWIzPBQB1MSLmbOxtoB1')


if __name__ == "__main__":
    unittest.main()



from mmisdk import CustodianFactory

# Instantiate the factory
factory = CustodianFactory()

# Create the custodian, using the factory
custodian = factory.create_for("qredo", "YOUR-REFRESH-TOKEN")

# Build tx details
qredo_tx_details = {
    "chainId": 1,
    "note": "test"
}
tx_params = {
    "from_": "0x62468FD916bF27A3b76d3de2e5280e53078e13E1",
    "to": "0x62468FD916bF27A3b76d3de2e5280e53078e13E1",
    "value": "1",
    "gasLimit": "21000",
    "gasPrice": "1000",
    "data": "",
    # "type": "2"
    # "maxPriorityFeePerGas": "12321321",
    # "maxFeePerGas": "12321321",
}

# Create the tx from details and send it to the custodian
transaction = custodian.create_transaction(qredo_tx_details, tx_params)
print(transaction)

# Get the created transaction
result = custodian.get_transaction(transaction.id)
print(result)

# {
#   "id": "1yiq2w7NSFei1uyQOremLiwbP2I",
#   "type": "1",
#   "from_": "0x6cF8426Cf36B1f43ffF6C376d290114297441bc2",
#   "to": "0x722dd3F80BAC40c951b51BdD28Dd19d435762180",
#   "value": "0",
#   "gas": "34062",
#   "gasPrice": "1000000010",
#   "maxPriorityFeePerGas": null,
#   "maxFeePerGas": null,
#   "nonce": "3",
#   "data": "0x97c5ed1e....",
#   "hash": "0xf31f4aa0d29e9d52a0f6de6dbb191f6060fd8bc994cea5c8785b2a94301a140c",
#   "status": {
#     "finished": false,
#     "submitted": true,
#     "signed": true,
#     "success": false,
#     "displayText": "Confirmed",
#     "reason": "Unknown"
#   }
# }



def map_transaction_status(status: str, reason: str):
    if (status == 'pending'):
        return {
            "finished": False,
            "submitted": False,
            "signed": False,
            "success": False,
            "displayText": 'Pending',
            "reason": reason,
        }
    if (status == 'created'):
        return {
            "finished": False,
            "submitted": False,
            "signed": False,
            "success": False,
            "displayText": 'Created',
            "reason": reason,
        }
    if (status == 'authorized'):
        return {
            "finished": False,
            "submitted": False,
            "signed": False,
            "success": False,
            "displayText": 'Authorized',
            "reason": reason,
        }
    if (status == 'approved'):
        return {
            "finished": False,
            "submitted": False,
            "signed": False,
            "success": False,
            "displayText": 'Approved',
            "reason": reason,
        }
    elif (status == 'signed'):
        return {
            "finished": False,
            "submitted": False,
            "signed": True,
            "success": False,
            "displayText": 'Signed',
            "reason": reason,
        }
    elif (status == 'scheduled'):
        return {
            "finished": False,
            "submitted": True,
            "signed": True,
            "success": False,
            "displayText": 'Scheduled',
            "reason": reason,
        }
    elif (status == 'pushed'):
        return {
            "finished": False,
            "submitted": True,
            "signed": True,
            "success": False,
            "displayText": 'Pushed',
            "reason": reason,
        }
    elif (status == 'confirmed'):
        return {
            "finished": False,
            "submitted": True,
            "signed": True,
            "success": False,
            "displayText": 'Confirmed',
            "reason": reason,
        }
    elif (status == 'mined'):
        return {
            "finished": True,
            "submitted": True,
            "signed": True,
            "success": True,
            "displayText": 'Mined',
            "reason": reason,
        }
    elif (status == 'failed'):
        return {
            "finished": True,
            "submitted": True,
            "signed": True,
            "success": False,
            "displayText": 'Failed',
            "reason": reason,
        }
    elif (status == 'expired'):
        return {
            "finished": True,
            "submitted": False,
            "signed": False,
            "success": False,
            "displayText": 'Aborted',
            "reason": reason,
        }
    elif (status == 'cancelled'):
        return {
            "finished": True,
            "submitted": False,
            "signed": False,
            "success": False,
            "displayText": 'Cancelled',
            "reason": reason,
        }
    elif (status == 'rejected'):
        return {
            "finished": True,
            "submitted": False,
            "signed": False,
            "success": False,
            "displayText": 'Rejected',
            "reason": reason,
        }
    else:
        return {
            "finished": False,
            "submitted": False,
            "signed": False,
            "success": False,
            "displayText": 'Unknown',
            "reason": reason,
        }



from abc import abstractmethod

from mmisdk.common.transaction import Transaction


class Custodian:
    """Generic class that each custodian client must inherit from."""

    def __init__(self, name, api_url, refresh_token):
        self.name = name
        self.api_url = api_url
        self.refresh_token = refresh_token

    @abstractmethod
    def get_transaction(self, transaction_id) -> Transaction:
        """Query the custodian's API to fetch details about the transaction of passed id."""
        pass

    @abstractmethod
    def create_transaction(self, tx_params) -> Transaction:
        """Create a transaction from the passed parameters a send it to the custodian's API."""
        pass



from typing import Dict
from typing import List
from typing import Optional

from pydantic import BaseModel


class QredoTransactionEvent(BaseModel):
    id: str
    timestamp: int
    status: str
    message: str


class QredoTransaction(BaseModel):
    txID: str
    txHash: str
    status: str
    timestamps: Dict
    events: List[QredoTransactionEvent]
    nonce: int
    gasLimit: str
    gasPrice: Optional[str]
    maxFeePerGas: Optional[str]
    maxPriorityFeePerGas: Optional[str]
    from_: str
    to: str
    value: str
    data: str
    rawTX: str
    createdBy: str
    accountID: str
    network: str
    chainID: str
